#!/usr/bin/env ruby
#
# Copyright (c) 2004 SASADA Koichi <ko1 at atdot.net>
#
# This program is free software with ABSOLUTELY NO WARRANTY.
# You can re-distribute and/or modify this program under
# the same terms of the Ruby's lisence.
#
#
# $Id: ndk_config.rb,v 1.31 2004/04/23 11:30:41 ko1 Exp $
# Create : K.S. 04/04/17 16:50:33
#
#
# You can check RCFILE with following command:
#
#   ruby ndk_config.rb [RCFILE]
#

require 'uri'
require 'socket'
require 'drb/acl'

module Nadoka
  class NDK_Logger
    def initialize manager, config
      @manager = manager
      @config  = config
      @dlog    = @config.debug_log
    end
    
    def dlog msg
      if @config.loglevel >= 3
        logging(@dlog, msg)
      end
    end
    
    def slog msg
      if @config.loglevel >= 2
        logging(make_logfilename(@config.system_log), "[NDK] #{msg}")
      end
    end
    
    def clog ch, msg
      ch = canonical_channel_name(ch)
      
      logfile = (@config.channel_info[ch] && @config.channel_info[ch][:log]) ||
                 @config.default_log
      logfile = make_logfilename(logfile, ch)
      
      logging(logfile, msg)
      
      @manager.state.backlog_push "tail: #{Time.now.strftime(@config.log_timeformat)} #{msg}"
    end

    def log msg
      if @config.loglevel >= 1
        logging(make_logfilename(@config.system_log), msg)
      end
    end

    def logging(io, msg)
      msg = "#{Time.now.strftime(@config.log_timeformat)} #{msg}"
      if io.respond_to? :puts
        io.puts msg
      else
        bdir = File.expand_path(@config.log_dir)
        open(File.expand_path(io, bdir), 'a'){|f|
          f.puts msg
        }
      end
    end

    def canonical_channel_name ch
      ch = ch.sub(/^\!.{5}/, '!')
      ch.downcase
    end
    
    RName = {        # ('&','#','+','!')
      '#' => 'Cs-',
      '&' => 'Ca-',
      '+' => 'Cp-',
      '!' => 'Ce-',
    }
    def make_logfilename(tmpl, ch='')
      ch  = ch.sub(/^[\&\#\+\!]|/){|c|
        RName[c]
      }
      ch  = URI.encode(ch)
      ch  = ch.gsub(/\*|\:/, '_')
      ch  = ch.gsub(/\//, 'I')
      str = Time.now.strftime(tmpl)
      str.gsub!(/\$\{setting_name\}/, @config.setting_name)
      str.gsub!(/\$\{channel_name\}/, ch)
      str
    end
  end

  class NDK_ConfigBase
    # system
    # 0: quiet, 1: normal, 2: system, 3: debug
    Loglevel     = 2
    Setting_name = 'DefaultSetting',
    
    # client server
    Client_server_port = 6667
    Client_server_host = nil
    Client_server_pass = 'NadokaPassWord' # or nil
    Client_server_acl  = nil
    ACL_Object = nil
    
    # 
    Server_list = [{
      :host => '127.0.0.1',
      :port => 6667,
      :pass => nil
    }]
    Servers = []

    Default_channels   = []
    Login_channels     = []

    #
    User       = ENV['USER'] || ENV['USERNAME'] || 'nadokatest'
    Nick       = 'test_bot'
    Hostname   = Socket.gethostname
    Servername = '*'
    Realname   = 'test bot on nadoka'
    
    Away_Message = 'away'
    Away_Nick    = nil
    
    #
    Channel_info = {},
      
    # log
    Default_log = '${setting_name}-${channel_name}-%y%m%d.log'
    System_log  = '${setting_name}-system_log'
    Debug_log   = $stdout

    Backlog_lines = 20
    Log_TimeFormat= '%y/%m/%d-%H:%M:%S'
    
    # dirs
    Plugins_dir = './plugins'
    Log_dir     = './log'
    
    # bots
    BotFiles    = []
    BotConfig   = {}

    def self.inherited subklass
      ConfigClass << subklass
    end
  end
  ConfigClass = [NDK_ConfigBase]
  BotClass = []
  
  class NDK_Config
    NDK_ConfigBase.constants.each{|e|
      eval %Q{
        def #{e.downcase}
          @config['#{e.downcase}'.intern]
        end
      }
    }
    
    def initialize manager, rcfile = nil
      @manager = manager
      @bots = []
      load_config(rcfile || './nadokarc')
    end
    attr_reader :config, :bots

    def remove_previous_setting
      # remove setting class
      klass = ConfigClass.shift
      while k = ConfigClass.shift
        Object.module_eval{
          remove_const(k.name)
        }
      end
      ConfigClass.push(klass)

      # remove bot class
      while k = BotClass.shift
        Object.module_eval{
          remove_const(k.name)
        }
      end
      # destruct bot instances
      @bots.each{|bot|
        bot.bot_desctruct
      }
      
      GC.start
    end
    
    def load_config(rcfile)
      remove_previous_setting
      
      load(rcfile) if rcfile
      
      @config = {}
      klass = ConfigClass.last
      klass.constants.each{|e|
        @config[e.downcase.intern] = klass.const_get(e)
      }

      if $NDK_Debug
        @config[:loglevel] = 3
      end
      
      @logger = NDK_Logger.new(@manager, self)
      slog "load config: #{rcfile}"

      if svrs = klass.const_get(:Servers)
        svl = []
        svrs.each{|si|
          ports = si[:port] || 6667
          host  = si[:host]
          pass  = si[:pass]
          if ports.respond_to? :each
            ports.each{|port|
              svl << {:host => host, :port => port, :pass => pass}
            }
          else
            svl <<   {:host => host, :port => ports, :pass => pass}
          end
        }
        @config[:server_list] = svl
      end
      
      if chs = klass.const_get(:Channel_info)
        dchs = []
        lchs = []
        cchs = {}
        chs.each{|ch, setting|
          if setting[:timing] == :startup
            dchs << ch
          else
            lchs << ch
          end
          cchs[ch.downcase] = setting
        }
        chs.replace cchs
        @config[:default_channels] = dchs
        @config[:login_channels]   = lchs
      end

      # acl setting
      if @config[:client_server_acl] && !@config[:acl_object]
        acl = @config[:client_server_acl].split(/\s+/).map{|e|
          e.size == 0 ? nil : e
        }.compact
        @config[:acl_object] = ACL.new(acl)
        slog "ACL: #{acl.join(' ')}"
      end

      # load bots
      @config[:botfiles].each{|file|
        if @config[:plugins_dir].respond_to? :each
          @config[:plugins_dir].each{|dir|
            break if load_botfile "#{dir}/#{file}.nb"
          }
        else
          load_botfile "#{@config[:plugins_dir]}/#{file}.nb"
        end
      }
      
      @bots = BotClass.map{|bk|
        bot = bk.new @manager, self
        slog "bot instance: #{bot}"
        bot
      }
    end

    def load_botfile file
      if FileTest.exist? file
        load file
        true
      else
        false
      end
    end

    # debug message
    def dlog msg
      @logger.dlog msg
    end
    
    # system message
    def slog msg
      @logger.slog msg
    end

    # normal message
    def log msg
      @logger.log msg
    end
    
    # channel message
    def clog ch, msg
      @logger.clog ch, msg
    end
    
  end
end

if $0 == __FILE__
  require 'pp'
  pp Nadoka::NDK_Config.new(ARGV.shift)
end

